#!/usr/bin/env python3

import unittest
from mock import patch

from km3db import StreamDS, CLBMap
from km3db.tools import (
    clbupi2compassupi,
    tonamedtuples,
    show_compass_calibration,
    detx,
    detx_for_run,
    todetid,
    todetoid,
)
from km3net_testdata import data_path


class TestStreamDSOnline(unittest.TestCase):
    def setUp(self):
        self.sds = StreamDS()

    def test_print_streams(self):
        self.sds.print_streams()

    def test_get(self):
        detectors = self.sds.detectors()
        assert (
            "OID\tSERIALNUMBER\tLOCATIONID\tCITY\tFIRSTRUN\tLASTRUN\nD_DU1CPPM\t2\tA00070004"
            in detectors
        )

    def test_pandas_as_container(self):
        detectors = self.sds.detectors(container="pd")
        assert 35 <= len(detectors)

    def test_namedtuple_as_container(self):
        detectors = self.sds.detectors(container="nt")
        assert 35 <= len(detectors)
        assert detectors[2].oid == "D_DU2NAPO"


class TestStreamDSOffline(unittest.TestCase):
    @patch("km3db.core.DBManager")
    def setUp(self, db_manager_mock):
        with open(data_path("db/streamds_output.txt"), "r") as fobj:
            streamds_meta = fobj.read()
        db_manager_mock_obj = db_manager_mock.return_value
        db_manager_mock_obj.get.return_value = streamds_meta
        self.sds = StreamDS()

    def test_streams(self):
        assert len(self.sds.streams) == 30
        assert "ahrs" in self.sds.streams
        assert "clbmap" in self.sds.streams
        assert "datalogevents" in self.sds.streams
        assert "datalognumbers" in self.sds.streams

    def test_getattr(self):
        assert hasattr(self.sds, "t0sets")
        assert hasattr(self.sds, "vendorhv")

    def test_attr_are_callable(self):
        self.sds.runs()
        self.sds.runsetupparams()
        self.sds.upi()

    def test_mandatory_selectors(self):
        assert "-" == self.sds.streams["productloc"].mandatory_selectors
        assert "detid" == self.sds.streams["runs"].mandatory_selectors
        assert (
            "detid,minrun,maxrun"
            == self.sds.streams["datalognumbers"].mandatory_selectors
        )

    def test_optional_selectors(self):
        self.assertEqual(
            "upi,city,locationid,operation,operationid",
            self.sds.streams["productloc"].optional_selectors,
        )
        self.assertEqual(
            "run,runjobid,jobtarget,jobpriority",
            self.sds.streams["runs"].optional_selectors,
        )
        self.assertEqual(
            "source_name,parameter_name",
            self.sds.streams["datalognumbers"].optional_selectors,
        )

    def test_print_streams(self):
        self.sds.print_streams()


class TestCLBMapOffline(unittest.TestCase):
    @patch("km3db.tools.StreamDS")
    def setUp(self, streamds_mock):
        streamds_mock_obj = streamds_mock.return_value
        with open(data_path("db/clbmap.txt"), "r") as fobj:
            streamds_mock_obj.get.return_value = tonamedtuples(
                "CLB", fobj.read(), renamemap=CLBMap.renamemap
            )
        self.clbmap = CLBMap("a")

    def test_length(self):
        assert 57 == len(self.clbmap)

    def test_clb_by_upi(self):
        print(self.clbmap.upis)
        assert 806487231 == self.clbmap.upis["3.4.3.2/V2-2-1/2.570"].dom_id
        assert 808964852 == self.clbmap.upis["3.4.3.2/V2-2-1/2.100"].dom_id
        assert 808982547 == self.clbmap.upis["3.4.3.2/V2-2-1/2.121"].dom_id
        assert 808961480 == self.clbmap.upis["3.4.3.2/V2-2-1/2.94"].dom_id
        assert 13 == self.clbmap.upis["3.4.3.2/V2-2-1/2.570"].floor
        assert 3 == self.clbmap.upis["3.4.3.2/V2-2-1/2.100"].du
        assert 121 == self.clbmap.upis["3.4.3.2/V2-2-1/2.121"].serial_number
        assert "D_ORCA003" == self.clbmap.upis["3.4.3.2/V2-2-1/2.94"].det_oid
        for upi in self.clbmap.upis.keys():
            assert upi == self.clbmap.upis[upi].upi

    def test_clb_by_dom_id(self):
        assert "3.4.3.2/V2-2-1/2.570" == self.clbmap.dom_ids[806487231].upi
        assert "3.4.3.2/V2-2-1/2.100" == self.clbmap.dom_ids[808964852].upi
        assert "3.4.3.2/V2-2-1/2.121" == self.clbmap.dom_ids[808982547].upi
        assert "3.4.3.2/V2-2-1/2.94" == self.clbmap.dom_ids[808961480].upi
        assert 13 == self.clbmap.dom_ids[806487231].floor
        assert 3 == self.clbmap.dom_ids[808964852].du
        assert 121 == self.clbmap.dom_ids[808982547].serial_number
        assert "D_ORCA003" == self.clbmap.dom_ids[808961480].det_oid
        for dom_id in self.clbmap.dom_ids.keys():
            assert dom_id == self.clbmap.dom_ids[dom_id].dom_id

    def test_get_base(self):
        assert 0 == self.clbmap.base(1).floor
        assert 0 == self.clbmap.base(3).floor
        assert 0 == self.clbmap.base(4).floor
        assert 808476701 == self.clbmap.base(1).dom_id
        assert 808981515 == self.clbmap.base(3).dom_id
        assert 808967761 == self.clbmap.base(4).dom_id


class TestCLBUPI2CompassUPI(unittest.TestCase):
    def test_ahrs(self):
        assert "3.4.3.4/AHRS/1.551" == clbupi2compassupi("3.4.3.2/V2-2-1/2.551")
        assert "3.4.3.4/AHRS/1.76" == clbupi2compassupi("3.4.3.2/V2-2-1/2.76")

    def test_lsm(sefl):
        assert "3.4.3.4/LSM303/3.1106" == clbupi2compassupi("3.4.3.2/V2-2-1/3.1013")
        assert "3.4.3.4/LSM303/3.948" == clbupi2compassupi("3.4.3.2/V2-2-1/3.855")


class TestShowCompassCalibration(unittest.TestCase):
    def test_function(self):
        show_compass_calibration("3.4.3.2/V2-2-1/2.551")


class TestDetxHelpersOnline(unittest.TestCase):
    def test_detx(self):
        det_id = 49
        detx_string = detx(49)
        assert detx_string is not None
        split_lines = detx_string.split("\n")
        assert str(det_id) in split_lines[0]
        assert 3461 == len(split_lines)
        assert float(split_lines[6].split()[-1]) == 0.0  # t0 is not set

    def test_detx_for_run(self):
        det_id = 49
        detx_string = detx_for_run(49, 8220)
        assert detx_string is not None
        split_lines = detx_string.split("\n")
        assert str(det_id) in split_lines[0]
        assert 3461 == len(split_lines)
        assert float(split_lines[6].split()[-1]) != 0.0  # check a single PMT t0


class TestDetIDandOIDConversionsOnline(unittest.TestCase):
    def test_todetid(self):
        assert 49 == todetid("D_ORCA006")

    def test_todetoid(self):
        assert "D_ORCA006" == todetoid(49)



#!/usr/bin/env python3
from collections import OrderedDict, namedtuple
from functools import wraps
import io
import json

import numpy as np

import km3db.core
import km3db.extras
from km3db.logger import log


try:
    # Python 3.5+
    from inspect import Signature, Parameter

    SKIP_SIGNATURE_HINTS = False
except ImportError:
    # Python 2.7
    SKIP_SIGNATURE_HINTS = True


class StreamDS:
    """Access to the streamds data stored in the KM3NeT database.

    Parameters
    ==========
    url: str (optional)
      The URL of the database web API
    container: str or None (optional)
      The default containertype when returning data.
        None (default): the data, as returned from the DB
          "nt": `namedtuple`, can be used when no pandas is available
          "pd": `pandas.DataFrame`, as returned in KM3Pipe v8 and below
    """

    def __init__(self, url=None, container=None):
        self._db = km3db.core.DBManager(url=url)
        self._streams = None
        self._update_streams()
        self._default_container = container

    @property
    def streams(self):
        return self._streams

    def _update_streams(self):
        """Update the list of available straems"""
        content = self._db.get("streamds")
        self._streams = OrderedDict()
        for entry in tonamedtuples("Stream", content):
            self._streams[entry.stream] = entry
            setattr(self, entry.stream, self.__getattr__(entry.stream))

    def __getattr__(self, attr):
        """Magic getter which optionally populates the function signatures"""
        if attr in self.streams:
            stream = self.streams[attr]
        else:
            raise AttributeError

        def func(**kwargs):
            return self.get(attr, **kwargs)

        func.__doc__ = stream.description

        if not SKIP_SIGNATURE_HINTS:
            sig_dict = OrderedDict()
            for sel in stream.mandatory_selectors.split(","):
                if sel == "-":
                    continue
                sig_dict[Parameter(sel, Parameter.POSITIONAL_OR_KEYWORD)] = None
            for sel in stream.optional_selectors.split(","):
                if sel == "-":
                    continue
                sig_dict[Parameter(sel, Parameter.KEYWORD_ONLY)] = None
            func.__signature__ = Signature(parameters=sig_dict)

        return func

    def print_streams(self):
        """Print the documentation for all available streams."""
        for stream in sorted(self.streams.values()):
            self._print_stream_parameters(stream)

    def _print_stream_parameters(self, stream):
        """Print the documentation for a given stream."""
        print("{}".format(stream.stream))
        print("-" * len(stream.stream))
        print("{}".format(stream.description))
        print("  available formats:   {}".format(stream.formats))
        print("  mandatory selectors: {}".format(stream.mandatory_selectors))
        print("  optional selectors:  {}".format(stream.optional_selectors))
        print()

    def help(self, stream_name):
        """Print help for a given stream"""
        try:
            self._print_stream_parameters(self.streams[stream_name])
        except KeyError:
            log.error("There is no stream called '{}'".format(stream_name))
            print(
                "Available streams:\n{}".format(
                    ", ".join(s.stream for s in sorted(self.streams.values()))
                )
            )

    def get(self, stream, fmt="txt", container=None, renamemap=None, **kwargs):
        """Retrieve the data for a given stream manually

        Parameters
        ==========
        stream: str
          Name of the stream (e.g. detectors)
        fmt: str ("txt", "text", "bin")
          Retrieved raw data format, depends on the stream type
        container: str or None
          The container to wrap the returned data, as specified in
          `StreamDS`.
        """
        sel = "".join(["&{0}={1}".format(k, v) for (k, v) in kwargs.items()])
        url = "streamds/{0}.{1}?{2}".format(stream, fmt, sel[1:])
        log.debug("URL: %s" % url)
        data = self._db.get(url)
        if not data:
            log.error("No data found at URL '%s'." % url)
            return
        if data.startswith("ERROR"):
            log.error(data)
            return

        if container is None and self._default_container is not None:
            container = self._default_container

        try:
            if container == "pd":
                return topandas(data)
            if container == "nt":
                return tonamedtuples(stream.capitalize(), data, renamemap=renamemap)
        except ValueError:
            log.critical(
                "Unable to convert data to container type '{}'. "
                "Database response: {}".format(container, data)
            )
        else:
            return data


class JSONDS:
    """Access to the jsonds data stored in the KM3NeT database.

    Parameters
    ==========
    url: str (optional)
      The URL of the database web API

    """

    def __init__(self, url=None):
        self._db = km3db.core.DBManager(url=url)

    def get(self, url):
        "Get JSON-type content from the url"
        content = self._db.get("jsonds/" + url)
        try:
            json_content = json.loads(content.decode())
        except AttributeError:
            json_content = json.loads(content)
        if json_content.get("Comment") is not None:
            log.warning(json_content["Comment"])
        if json_content["Result"] != "OK":
            log.critical("Error from DB: %s", json_content.get("Data"))
            raise ValueError("Error while retrieving the parameter list.")
        return json_content["Data"]


class CLBMap:
    renamemap = dict(
        DETOID="det_oid",
        UPI="upi",
        DOMID="dom_id",
        DUID="du",
        SERIALNUMBER="serial_number",
        FLOORID="floor",
    )

    def __init__(self, det_oid):
        # if isinstance(det_oid, numbers.Integral):
        #     db = km3db.core.DBManager()
        #     # det_oid and det_id chaos in the database
        #     # _det_oid = db.get_det_oid(det_oid)
        #     # if _det_oid is not None:
        #     #     det_oid = _det_oid
        self.det_oid = det_oid
        sds = StreamDS(container="nt")
        self._data = sds.get("clbmap", detoid=det_oid, renamemap=self.renamemap)
        self._by = {}

    def __len__(self):
        return len(self._data)

    @property
    def upis(self):
        """A dict of CLBs with UPI as key"""
        parameter = "upi"
        if parameter not in self._by:
            self._populate(by=parameter)
        return self._by[parameter]

    @property
    def dom_ids(self):
        """A dict of CLBs with DOM ID as key"""
        parameter = "dom_id"
        if parameter not in self._by:
            self._populate(by=parameter)
        return self._by[parameter]

    @property
    def omkeys(self):
        """A dict of CLBs with the OMKey tuple (DU, floor) as key"""
        parameter = "omkey"
        if parameter not in self._by:
            self._by[parameter] = {}
            for clb in self.upis.values():
                omkey = (clb.du, clb.floor)
                self._by[parameter][omkey] = clb
            pass
        return self._by[parameter]

    def base(self, du):
        """Return the base CLB for a given DU"""
        parameter = "base"
        if parameter not in self._by:
            self._by[parameter] = {}
            for clb in self._data:
                if clb.floor == 0:
                    self._by[parameter][clb.du] = clb
        return self._by[parameter][du]

    def _populate(self, by):
        data = {}
        for clb in self._data:
            data[getattr(clb, by)] = clb
        self._by[by] = data


def lru_cache(func):
    """Poor mans lru_cache for compatiblity"""
    cache = {}

    @wraps(func)
    def wrapper(*args):
        key = tuple(args)
        if key not in cache:
            cache[key] = func(*args)
        return cache[key]

    return wrapper


@lru_cache
def clbupi2compassupi(clb_upi):
    """Return Compass UPI from CLB UPI."""
    sds = StreamDS(container="nt")
    upis = [i.content_upi for i in sds.integration(container_upi=clb_upi)]
    compass_upis = [upi for upi in upis if ("AHRS" in upi) or ("LSM303" in upi)]
    if len(compass_upis) > 1:
        log.warning(
            "Multiple compass UPIs found for CLB UPI {}. "
            "Using the first entry.".format(clb_upi)
        )
    return compass_upis[0]


@lru_cache
def todetoid(det_id):
    """Convert det OID (e.g. D_ORCA006) to det ID (e.g. 49)

    If a det OID is provided it will simple be returned.
    """
    try:
        det_id = int(det_id)
    except ValueError:
        # assume it's an OID
        return det_id

    detectors = StreamDS(container="nt").get("detectors")
    for detector in detectors:
        if detector.serialnumber == det_id:
            return detector.oid
    log.error("No detector with det ID '{}' found to look up its OID".format(det_id))


@lru_cache
def todetid(det_oid):
    """Convert det ID (e.g. 49) to det OID (e.g. D_ORCA006)

    If a det OID is provided it will simple be returned.
    """
    if isinstance(det_oid, int):
        return det_oid
    detectors = StreamDS(container="nt").get("detectors")
    for detector in detectors:
        if detector.oid == det_oid:
            return detector.serialnumber
    log.error("Could not convert det OID '{}' to ID".format(det_oid))


def tonum(value):
    """Convert a value to a numerical one if possible"""
    for converter in (int, float):
        try:
            return converter(value)
        except (ValueError, TypeError):
            pass
    return value


def tonamedtuples(name, text, renamemap=None):
    """Creates a list of namedtuples from database output

    Parameters
    ----------
    name: str
      Name of the namedtuple
    text: str
      Raw output from the database (tab separated values
      and the first line being the header)
    renamemap: dict(str: str) or None (default)
      Rename the fields according to this map.
    """
    if renamemap is None:
        renamemap = {}
    lines = text.split("\n")
    cls = namedtuple(name, [renamemap.get(s, s.lower()) for s in lines.pop(0).split()])
    entries = []
    for line in lines:
        if not line:
            continue
        entries.append(cls(*map(tonum, line.split("\t"))))
    return entries


def topandas(text):
    """Create a DataFrame from database output"""
    return km3db.extras.pandas().read_csv(io.StringIO(text), sep="\t")


def df_to_sarray(df):
    """
    Convert a pandas DataFrame object to a numpy structured array.

    Parameters
    ----------
    df : Pandas.DataFrame
      the data frame to convert

    Returns
    -------
    A numpy structured array representation of df.
    """

    for dtype in df.dtypes:
        if dtype is np.dtype("O"):
            log.critical(
                "At least one column contains strings, "
                "which are currently not supported in the HDF5 backend. "
                "The CSV backend will work fine."
            )
            exit(1)

    cols = df.columns
    types = [(cols[i], df[k].dtype.type) for (i, k) in enumerate(cols)]
    v = df.values
    arr = np.zeros(v.shape[0], np.dtype(types))
    for (idx, field) in enumerate(arr.dtype.names):
        arr[field] = v[:, idx]
    return arr


def show_compass_calibration(clb_upi, version="3"):
    """Show compass calibration data for given `clb_upi`."""
    db = km3db.core.DBManager()
    compass_upi = clbupi2compassupi(clb_upi)
    compass_model = compass_upi.split("/")[1]
    print("Compass UPI: {}".format(compass_upi))
    print("Compass model: {}".format(compass_model))
    content = db.get(
        "show_product_test.htm?upi={}&"
        "testtype={}-CALIBRATION-v{}&n=1&out=xml".format(
            compass_upi, compass_model, version
        )
    ).replace("\n", "")

    import xml.etree.ElementTree as ET

    try:
        root = ET.parse(io.StringIO(content)).getroot()
    except ET.ParseError:
        print("No calibration data found")
    else:
        for child in root:
            print("{}: {}".format(child.tag, child.text))
        names = [c.text for c in root.findall(".//Name")]
        values = [[i.text for i in c] for c in root.findall(".//Values")]
        for name, value in zip(names, values):
            print("{}: {}".format(name, value))


def detx(det_id, pcal=0, rcal=0, tcal=0, version=2):
    """Retrieve the calibrated detector file for the given detector ID"""

    url = "detx/{det_id}?tcal={tcal}&pcal={pcal}&rcal={rcal}&v={version}".format(
        det_id=det_id, tcal=tcal, pcal=pcal, rcal=rcal, version=version
    )

    return km3db.core.DBManager().get(url)


def detx_for_run(det_id, run, version=2):
    """Retrieve the calibrate detector file for given run"""
    run_table = StreamDS(container="nt").get("runs", detid=det_id)
    if run_table is None:
        log.error("No run table found for detector ID {}".format(det_id))
        return None

    for run_info in run_table:
        if run_info.run == run:
            break
    else:
        log.error("Run {} not found for detector {}".format(run, det_id))
        return None

    tcal = run_info.t0_calibsetid
    if tcal == "":
        log.warning(
            "No time calibration found for run {} (detector {})".format(run, det_id)
        )
        tcal = 0

    pcal = run_info.pos_calibsetid
    if pcal == "":
        log.warning(
            "No position calibration found for run {} (detector {})".format(run, det_id)
        )
        pcal = 0

    rcal = run_info.rot_calibsetid
    if rcal == "":
        log.warning(
            "No rotation calibration found for run {} (detector {})".format(run, det_id)
        )
        rcal = 0

    return detx(det_id, pcal=pcal, rcal=rcal, tcal=tcal, version=version)



#!/usr/bin/env python3
"""
Access the KM3NeT StreamDS DataBase service.

Usage:
    streamds
    streamds list
    streamds info STREAM
    streamds get [-f FORMAT -o OUTFILE -g COLUMN] STREAM [PARAMETERS...]
    streamds upload [-q -x] CSV_FILE
    streamds (-h | --help)
    streamds --version

Options:
    STREAM      Name of the stream.
    PARAMETERS  List of parameters separated by space (e.g. detid=29).
    CSV_FILE    Whitespace separated data for the runsummary tables.
    -f FORMAT   Usually 'txt' for ASCII or 'text' for UTF-8 [default: txt].
    -o OUTFILE  Output file: supported formats '.csv' and '.h5'.
    -g COLUMN   Group dataset by the name of the given row when writing HDF5.
    -q          Test run! When uploading, a TEST_ prefix will be added to the data.
    -x          Do not verify the SSL certificate.
    -h --help   Show this screen.

"""
import getpass
import json
import logging
import os
import requests

import km3db
import km3db.extras
from docopt import docopt

log = km3db.logger.get_logger("streamds")

RUNSUMMARYNUMBERS_URL = "https://km3netdbweb.in2p3.fr/jsonds/runsummarynumbers/i"
RUNSUMMARYSTRINGS_URL = "https://km3netdbweb.in2p3.fr/jsonds/runsummarystrings/i"
RUNSUMMARYSTRINGS_COLUMNS = set(["UUID", "JPP"])
REQUIRED_COLUMNS = set(["run", "det_id", "source"])
COLUMN_MAPPINGS = {"GIT": "source", "detector": "det_id"}


def print_streams():
    """Print all available streams with their full description"""
    sds = km3db.StreamDS()
    sds.print_streams()


def print_info(stream):
    """Print the information about a stream"""
    sds = km3db.StreamDS()
    sds.help(stream)


def get_data(stream, parameters, fmt, outfile=None, groupby=None):
    """Retrieve data for given stream and parameters, or None if not found"""
    sds = km3db.StreamDS()
    if stream not in sds.streams:
        log.error("Stream '{}' not found in the database.".format(stream))
        return
    params = {}
    if parameters:
        for parameter in parameters:
            if "=" not in parameter:
                log.error(
                    "Invalid parameter syntax '{}'\n"
                    "The correct syntax is 'parameter=value'".format(parameter)
                )
                continue
            key, value = parameter.split("=")
            params[key] = value
    data = sds.get(stream, fmt, **params)
    if data is not None:
        if outfile is not None:
            write_output(outfile, stream, data, groupby)
        else:
            try:
                print(data)
            except BrokenPipeError:
                pass
    else:
        sds.help(stream)


def write_output(outfile, stream, data, groupby=None):
    """Writes the DB output to a file (HDF5 or CSV)"""
    _, ext = os.path.splitext(outfile)

    if ext == ".h5":
        write_output_hdf5(outfile, stream, data, groupby=groupby)
        exit(0)
    if ext == ".csv":
        write_output_csv(outfile, stream, data)
        exit(0)

    log.error("Unsupported filetype with '{}'".format(ext))
    exit(1)


def write_output_hdf5(outfile, stream, data, groupby):
    """Write DB output to HDF5"""
    h5py = km3db.extras.h5py()
    df = km3db.tools.topandas(data)
    with h5py.File(outfile, "a") as h5f:
        if groupby is not None:
            for group, _df in df.groupby(groupby):
                sa = km3db.tools.df_to_sarray(_df)
                dset_name = stream + "/{}".format(group)
                if dset_name in h5f:
                    log.warning(
                        "Dataset '{}' already exists, skipping...".format(dset_name)
                    )
                    continue
                h5f.create_dataset(
                    stream + "/{}".format(group),
                    compression="gzip",
                    compression_opts=3,
                    data=sa,
                )
        else:
            sa = km3db.tools.df_to_sarray(df)
            h5f[stream] = sa
    print("Database output written to '{}'.".format(outfile))


def write_output_csv(outfile, stream, data):
    """Write DB output to CSV"""
    with open(outfile, "w") as fobj:
        fobj.write(data)
    print("Database output written to '{}'.".format(outfile))


def available_streams():
    """Show a short list of available streams."""
    sds = km3db.StreamDS()
    print("Available streams: ")
    print(", ".join(sorted(sds.streams)))


def upload_runsummary(csv_filename, testrun=False, verify=False):
    """Reads the CSV file and uploads its contents to the runsummary table"""
    pd = km3db.extras.pandas()

    print("Checking '{}' for consistency.".format(csv_filename))
    if not os.path.exists(csv_filename):
        log.critical("{} -> file not found.".format(csv_filename))
        return
    try:
        df = pd.read_csv(csv_filename, delim_whitespace=True)
    except pd.errors.EmptyDataError as e:
        log.error(e)
        return

    for fromcol, tocol in COLUMN_MAPPINGS.items():
        if fromcol in df.columns:
            log.warn("Renaming column '{}' to '{}'.".format(fromcol, tocol))
            df[tocol] = df[fromcol]
            df.drop(columns=fromcol, inplace=True)

    cols = set(df.columns)

    if not REQUIRED_COLUMNS.issubset(cols):
        log.error(
            "Missing columns: {}.".format(
                ", ".join(str(c) for c in REQUIRED_COLUMNS - cols)
            )
        )
        return

    parameters = cols - REQUIRED_COLUMNS
    if len(parameters) < 1:
        log.error("No parameter columns found.")
        return

    if len(df) == 0:
        log.critical("Empty dataset.")
        return

    print(
        "Found data for parameters: {}.".format(", ".join(str(c) for c in parameters))
    )
    print("Converting CSV data into JSON")
    if testrun:
        log.warn("Test run: adding 'TEST_' prefix to parameter names")
        prefix = "TEST_"
    else:
        prefix = ""

    det_id_zero_mask = df["det_id"] == 0
    if sum(det_id_zero_mask) > 0:
        log.warning("Entries with 'det_id=0' found, removing them.")
        df = df[~det_id_zero_mask]
    df["det_id"] = df["det_id"].apply(km3db.tools.todetoid)

    data_runsummarynumbers = convert_runsummary_to_json(
        df[df.columns.difference(RUNSUMMARYSTRINGS_COLUMNS)], prefix=prefix
    )
    print(
        "We have {:.3f} MB runsummarynumbers to upload.".format(
            len(data_runsummarynumbers) / 1024**2
        )
    )
    _database_upload(data_runsummarynumbers, verify)

    data_runsummarystrings = convert_runsummary_to_json(
        df[list(REQUIRED_COLUMNS.union(RUNSUMMARYSTRINGS_COLUMNS))],
        prefix=prefix,
        isrunsummarystrings=True,
    )
    print(
        "We have {:.3f} MB runsummarystrings to upload.".format(
            len(data_runsummarystrings) / 1024**2
        )
    )
    _database_upload(data_runsummarystrings, verify, isrunsummarystrings=True)


def _database_upload(data, verify=False, isrunsummarystrings=False):
    db = km3db.DBManager()  # noqa

    print("Requesting database session.")
    session_cookie = db.session_cookie

    print("Uploading the data to the database.")
    url = RUNSUMMARYSTRINGS_URL if isrunsummarystrings else RUNSUMMARYNUMBERS_URL
    print("URL: {}".format(url))
    r = requests.post(
        url, cookies={"sid": session_cookie}, files={"datafile": data}, verify=verify
    )

    if r.status_code == 200:
        log.debug("POST request status code: {}".format(r.status_code))
        print("Database response:")
        db_answer = json.loads(r.text)
        for key, value in db_answer.items():
            print("  -> {}: {}".format(key, value))
        if db_answer["Result"] == "OK":
            print("Upload successful.")
        else:
            log.critical("Something went wrong.")
    else:
        log.error("POST request status code: {}".format(r.status_code))
        log.critical("Something went wrong...")
        return


def convert_runsummary_to_json(
    df,
    comment="Uploaded via km3pipe.StreamDS",
    prefix="TEST_",
    isrunsummarystrings=False,
):
    """Convert a Pandas DataFrame with runsummary to JSON for DB upload"""
    data_field = []
    comment += ", by {}".format(getpass.getuser())
    for det_id, det_data in df.groupby("det_id"):
        runs_field = []
        data_field.append({"DetectorId": det_id, "Runs": runs_field})

        for run, run_data in det_data.groupby("run"):
            parameters_field = []
            runs_field.append({"Run": int(run), "Parameters": parameters_field})

            parameter_dict = {}
            for row in run_data.iterrows():
                for parameter_name in run_data.columns:
                    if parameter_name in REQUIRED_COLUMNS:
                        continue

                    if parameter_name not in parameter_dict:
                        entry = {"Name": prefix + parameter_name, "Data": []}
                        parameter_dict[parameter_name] = entry
                    data_value = getattr(row[1], parameter_name)
                    if not isrunsummarystrings:
                        try:
                            data_value = float(data_value)
                        except ValueError as e:
                            log.critical("Data values has to be floats!")
                            raise ValueError(e)
                    else:
                        data_value = str(data_value)
                    value = {"S": str(getattr(row[1], "source")), "D": data_value}
                    parameter_dict[parameter_name]["Data"].append(value)
            for parameter_data in parameter_dict.values():
                parameters_field.append(parameter_data)
    data_to_upload = {"Comment": comment, "Data": data_field}
    file_data_to_upload = json.dumps(data_to_upload)
    return file_data_to_upload


def main():
    args = docopt(__doc__)

    if args["info"]:
        print_info(args["STREAM"])
    elif args["list"]:
        print_streams()
    elif args["upload"]:
        upload_runsummary(args["CSV_FILE"], args["-q"], args["-x"])
    elif args["get"]:
        get_data(
            args["STREAM"],
            args["PARAMETERS"],
            fmt=args["-f"],
            outfile=args["-o"],
            groupby=args["-g"],
        )
    else:
        available_streams()



#!/usr/bin/env python3
"""
Generate a cookie for the KM3NeT Oracle Web API.

Usage:
    km3dbcookie [-B | -C]
    km3dbcookie (-h | --help)
    km3dbcookie --version

Options:
    -B             Request the cookie for a class B network (12.23.X.Y).
    -C             Request the cookie for a class C network (12.23.45.Y).
    -h --help   Show this screen.

Example:

    $ km3dbcookie -B
    Please enter your KM3NeT DB username: tgal
    Password:
    Cookie saved as '/Users/tamasgal/.km3netdb_cookie'
    $ cat /Users/tamasgal/.km3netdb_cookie
    .in2p3.fr	TRUE	/	TRUE	0	sid	_tgal_131.188_70b78042c03a434594b041073484ce23


"""
import km3db
from docopt import docopt


def main():
    args = docopt(__doc__, version=km3db.version)
    if args["-B"]:
        db = km3db.DBManager(network_class="B")
    elif args["-C"]:
        db = km3db.DBManager(network_class="C")
    else:
        db = km3db.DBManager()
    db.request_session_cookie()

    print("Cookie saved as '{}'".format(km3db.core.COOKIE_FILENAME))



"""
Access run summary numbers stream
=================================

The following example shows how to use the run summary numbers stream
to access summary information per run.

- The first part shows the content of the Stream.
- The second part explore the usage to get detector-related summary
  number, like the overall HRV fraction or the DAQ event rate.
- The third part shows how to access DOM related information, like the
  average rate of a specific PMT during the run.

author: Valentin Pestel (vpestel@nikhef.nl)
"""

import km3db
import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

#####################################################
# First, we declare the sds client. This object will take care of the
# DB connection. ``container`` is set to `pd`, which means the sds
# will return ``pandas.DataFrame``.

sds = km3db.tools.StreamDS(container="pd")

#####################################################
# What contains runsummarynumbers stream ?
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To explore what contains this Stream, we will first query few runs
# from ORCA6 and print the resulting dataframe.

df = sds.runsummarynumbers(detid="D_ORCA006", minrun=8000, maxrun=8000)
print(df)

#####################################################
# Let's take a look in the `SOURCE_NAME` column :

print(df["SOURCE_NAME"].unique())

#####################################################
# We can already see 2 categories of `SOURCE_NAME':
#
# - Some corresponding to DOM module ID. These values are representing
#   summary information per DOM.
# - Some corresponding to version numbers. These values are
#   representing summary information for the whole detector. The
#   version number correspond to the JPP version used to compute these
#   meta-variables.


#####################################################
# Exploit detector-related summary information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# This time we will get a much larger set of runs, but we will only
# get the information from `SOURCE_NAME == 14.4.1` (last JPP version
# when this example was wrote).

run_min = 8000
run_max = 8500
detid = "D_ORCA006"

df = sds.runsummarynumbers(
    detid=detid, minrun=run_min, maxrun=run_max, source_name="14.4.1"
)

#####################################################
#
# We will use the `pivot` function to re-arrange the dataframe
# shape in a more practical way.

df = df.pivot(index="RUN", columns="PARAMETER_NAME", values="DATA_VALUE")
print(df)

#####################################################
#
# Now let's look at what is available in the columns :
print(list(df.columns))

#####################################################
#
# One can notice two time-related variables: `UTCMin_s` and
# `UTCMax_s`.  These two variables contains the start and stop time of
# the run, in seconds. In python, it's often more convenient to
# convert the time object to datetime-objects. Pandas is having a very
# convenient function for that: ``pandas.to_datetime``.  In the
# following lines, we are defining 3 new columns, respectively
# containing the start, mean and stop time of the run:

df["t_start"] = pd.to_datetime(df["UTCMin_s"], unit="s")
df["t_stop"] = pd.to_datetime(df["UTCMax_s"], unit="s")
df["t_mean"] = pd.to_datetime(np.mean(df[["UTCMin_s", "UTCMax_s"]], axis=1), unit="s")

print(df)

#####################################################
#
# As a first example, we can look at the evolution of HRV and mean PMT
# rate in function of time.  For the mean PMT rate, we also represent
# its +/- standard variation.

fig, axe = plt.subplots(figsize=[12, 4])

axeb = axe.twinx()

axe.plot(df["t_start"], df["MEAN_Rate_Hz"], color="C0")
ymin = df["MEAN_Rate_Hz"] - df["RMS_Rate_Hz"]
ymax = df["MEAN_Rate_Hz"] + df["RMS_Rate_Hz"]
axe.fill_between(df["t_start"], ymin, ymax, color="C0", alpha=0.5)

axeb.plot(df["t_start"], df["HRV"], color="C1")

axe.set_ylabel("PMT rate [kHz]")
axe.yaxis.label.set_color("C0")

axeb.set_ylabel("High rate veto fraction")
axeb.yaxis.label.set_color("C1")

#####################################################
#
# We can also look at the events rate, with a break down of the
# different trigger rates

triggers_name = ["JTrigger3DMuon", "JTrigger3DShower", "JTriggerMXShower"]

fig, axe = plt.subplots(figsize=[12, 4])
axe.set_yscale("log")

axe.plot(
    df["t_start"],
    df["JDAQEvent"] / df["livetime_s"],
    color="gray",
    label="DAQ events",
    zorder=0,
)

for name in triggers_name:
    axe.plot(df["t_start"], df[name] / df["livetime_s"], label=name, zorder=1)

axe.set_ylabel("Rate [kHz]")
axe.legend()

plt.tight_layout()


#####################################################
# Exploit DOM-related summary information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# First, we will remove the detector-related information from the
# dataframe. For this, we only keep the `SOURCE_NAME` starting by
# `80`. As all the `SOURCE_NAME` are now an integer, we convert them
# to int.

df = sds.runsummarynumbers(detid="D_ORCA006", minrun=8000, maxrun=8050)
df = df[df["SOURCE_NAME"].str.contains("^80.*")]
df = df.astype({"SOURCE_NAME": int})

#####################################################
# We will use the `pivot` function to re-arrange the dataframe
# shape in a more practical way.
# We would like a dataframe with the columns name being
# `PARAMETER_NAME`, each rows being the showing the `VALUES` for a
# given module (`SOURCE_NAME`) during a given `RUN`.
# This is done with the following line :

df = df.pivot(
    index=["SOURCE_NAME", "RUN"], columns="PARAMETER_NAME", values="DATA_VALUE"
)
print(df)

#####################################################
# We can look at the list of available columns :

print(list(df.columns))

#####################################################
# Finally, let's plot the evolution of the rate per PMT for a given
# module ID. Here, we randomly pick `806451572`.

mID = 806451572

fig, axe = plt.subplots(figsize=[8, 4])
axe.set_yscale("log")

cmap = mpl.cm.get_cmap("viridis")

for i in range(31):
    colName = "pmt_{}_mean_rate".format(i)
    color = cmap(i / 30.0)
    axe.plot(df.loc[mID].index, df.loc[mID][colName], color=color)

axe.set_xlabel("Run ID")
axe.set_ylabel("Mean rate [kHz]")
plt.tight_layout()


plt.show()



#!/usr/bin/env python3
"""
Command line access to the KM3NeT DB web API.

Usage:
    km3db URL
    km3db (-h | --help)
    km3db --version

Options:
    URL         The URL, starting from the database website's root.
    -h --help   Show this screen.

Example:

    km3db "streamds/runs.txt?detid=D_ARCA003"

"""
import km3db
from docopt import docopt


def main():
    args = docopt(__doc__, version=km3db.version)
    db = km3db.DBManager()
    result = db.get(args["URL"])
    if result is not None:
        print(result)



#!/usr/bin/env python3
"""
Retrieves DETX files from the database.

Usage:
    detx [options] DET_ID
    detx [options] DET_ID RUN
    detx (-h | --help)
    detx --version

Options:
    DET_ID        The detector ID (e.g. 49)
    RUN           The run ID.
    -p P_CAL      Position calibration ID [default: 0].
    -r R_CAL      Rotation calibration ID [default: 0].
    -t T_CAL      Time calibration ID [default: 0].
    -v VERSION    DETX file format version [default: 2].
    -o OUT        Output folder or filename.
    -h --help     Show this screen.

Example:

    detx 49 8220  # retrieve the calibrated DETX for run 8220 of ORCA6

"""
import km3db
from km3db.logger import log
from docopt import docopt


def main():
    args = docopt(__doc__, version=km3db.version)

    try:
        det_id = int(args["DET_ID"])
    except ValueError:
        log.error("Please proivde a valid detector ID (e.g. 49).")
        return

    if args["RUN"] is not None:
        detx = km3db.tools.detx_for_run(
            det_id, int(args["RUN"]), version=int(args["-v"])
        )
    else:
        detx = km3db.tools.detx(
            det_id,
            pcal=args["-p"],
            rcal=args["-r"],
            tcal=args["-t"],
            version=int(args["-v"]),
        )

    if detx is None:
        log.error("No detx found.")
        return

    if args["-o"]:
        with open(args["-o"], "w") as fobj:
            fobj.write(detx)
    else:
        try:
            print(detx)
        except BrokenPipeError:
            pass



"""
Access detector IDs and run tables
==================================

This example shows how to retrieve a detector version, and the list of
runs for this detector configuration.

author: Valentin Pestel (vpestel@nikhef.nl)
"""


import km3db
import pandas as pd

#####################################################
# First, we declare the sds client. This object will take care of the
# DB connection. ``container`` is set to `pd`, which means the sds
# will return ``pandas.DataFrame``.

sds = km3db.tools.StreamDS(container="pd")


#####################################################
# Getting the detectors list
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The detectors list is available in the `detectors` stream.

print(sds.detectors())

#####################################################
# We can see that the list of detectors is quite long, because it
# includes also testing setup in several laboratories.
# The actual detectors, ORCA and ARCA, are respectively associated to
# France and Italy as `CITY`. We can use these selectors to narrow
# down the selection:

print(sds.detectors(city="France"))
print(sds.detectors(city="Italy"))

#####################################################
# Often, we want to look at a specific detector, for which we have the
# serial number, e.g. ORCA6 with the serial number 49. It is possible
# to use that as a selector too:

print(sds.detectors(serialnumber=49))

#####################################################
# Getting the run list for a given detector
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Now that we have access to the detector list, we can look at the
# list of runs for a given setup. This is contained in the `runs`
# stream. We will use the last detector from the ARCA list as an
# example. Note that `detid` is the `runs` tables is the `OID` from
# the detector stream.  Let's look at the list of run for this
# configuration:

# Get ARCA21 detector
det = sds.detectors(city="Italy", serialnumber=133).iloc[0]
print(det)

runs = sds.runs(detid=det["OID"])
print(runs)

#####################################################
# This table contains information relative to the job settings:

print(list(runs.columns))

#####################################################
# We can see that the start/stop time of the run is provide with
# `UNIXJOBSTART`/`UNIXJOBEND`. Using ``pandas.to_datetime``, we create
# new variable in datetime format, allowing an easier usage.

runs["datetime_start"] = pd.to_datetime(runs["UNIXJOBSTART"], unit="ms")
runs["datetime_stop"] = pd.to_datetime(runs["UNIXJOBEND"], unit="ms")

print(runs)


#####################################################
# Access individual run
# ~~~~~~~~~~~~~~~~~~~~~
#
# It is often necessary to access the information run per run, or for
# a set of runs. A good option for that is to use the indexing of dataframe.
# First, we set the `RUN` to be the index:

runs = runs.set_index("RUN")

#####################################################
# The first column is now containing the run number, and each row can
# be accessed with this index, using the `loc` getter:

run_id = runs.sample(1).index[0]  # Get a random run number from the frame
print(f"For this example, we will be using {run_id}.")

#####################################################
# Now, to access the information for this run:
print(runs.loc[run_id])

#####################################################
# And accessing a specific information for this run:
print(f"The run setup name for run {run_id} was: {runs.loc[run_id]['RUNSETUPNAME']}")



import unittest
import mock
import tempfile

from km3db import DBManager
from km3db.core import on_whitelisted_host, SESSION_COOKIES, AuthenticationError


class TestKM3DB(unittest.TestCase):
    def test_init(self):
        DBManager()

    def test_whitelisted_hosts(self):
        for host, cookie in SESSION_COOKIES.items():
            if on_whitelisted_host(host):
                assert DBManager().session_cookie == cookie

    def test_get(self):
        db = DBManager()
        result = db.get("streamds/detectors.txt")
        assert result.startswith(
            "OID\tSERIALNUMBER\tLOCATIONID\tCITY\tFIRSTRUN\tLASTRUN\nD_DU1CPPM\t2\tA00070004\tMarseille"
        )

    @mock.patch("os.path.exists")
    @mock.patch("os.getenv")
    @mock.patch("urllib.request.urlopen")
    def test_request_session_cookie_from_env_with_credentials(
        self, urlopen_mock, getenv_mock, exists_mock
    ):
        class StreamMock:
            def read(self):
                return b"foo"

        def getenv_mock_side_effect(key):
            if key == "KM3NET_DB_USERNAME":
                return "username"
            if key == "KM3NET_DB_PASSWORD":
                return "password"
            return None

        urlopen_mock.return_value = StreamMock()
        exists_mock.return_value = False
        getenv_mock.side_effect = getenv_mock_side_effect

        db = DBManager()
        with self.assertRaises(AuthenticationError):
            cookie = db.request_session_cookie()

        getenv_mock.assert_has_calls(
            [
                mock.call("KM3NET_DB_USERNAME"),
                mock.call("KM3NET_DB_PASSWORD"),
            ]
        )

        urlopen_mock.assert_called_with(
            "https://km3netdbweb.in2p3.fr/home.htm?usr=username&pwd=password&persist=y"
        )

    @mock.patch("os.path.exists")
    @mock.patch("os.getenv")
    def test_request_session_cookie_from_env_with_cookie(
        self, getenv_mock, exists_mock
    ):
        class StreamMock:
            def read(self):
                return b"foo"

        the_cookie = "namnam"

        def getenv_mock_side_effect(key):
            if key == "KM3NET_DB_COOKIE":
                return the_cookie
            return None

        exists_mock.return_value = False
        getenv_mock.side_effect = getenv_mock_side_effect

        db = DBManager()

        cookie = db._request_session_cookie()

        getenv_mock.assert_has_calls(
            [
                mock.call("KM3NET_DB_COOKIE"),
            ]
        )

        assert the_cookie == cookie

    @mock.patch("os.path.exists")
    @mock.patch("os.getenv")
    def test_request_session_cookie_from_env_with_cookie_file(
        self, getenv_mock, exists_mock
    ):
        class StreamMock:
            def read(self):
                return b"foo"

        f = tempfile.NamedTemporaryFile(delete=False)
        f.write(b"ignoredstring namnam")
        f.close()

        def getenv_mock_side_effect(key):
            if key == "KM3NET_DB_COOKIE_FILE":
                return f.name
            return None

        exists_mock.return_value = False
        getenv_mock.side_effect = getenv_mock_side_effect

        db = DBManager()

        cookie = db._request_session_cookie()

        getenv_mock.assert_has_calls(
            [
                mock.call("KM3NET_DB_COOKIE_FILE"),
            ]
        )

        assert "namnam" == cookie



# Filename: runtable.py
"""
Prints the run table for a given detector ID.

Usage:
    runtable [options] DET_ID
    runtable (-h | --help)
    runtable --version

Options:
    -h --help           Show this screen.
    -c                  Compact view.
    -n RUNS             Number of runs.
    -r FROM_RUN-TO_RUN  Range of runs (example: 3100-3200).
    -t TARGET           Job target (run/on/off)
    -s REGEX            Regular expression to filter the runsetup name/id.
    DET_ID              Detector ID (eg. D_ARCA001).

"""

import re
import sys

import km3db
from km3db.logger import log

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def runtable(
    det_id, n=5, run_range=None, target=None, compact=False, sep="\t", regex=None
):
    """Print the run table of the last `n` runs for given detector"""
    runs = km3db.StreamDS(container="nt").get("runs", detid=det_id)

    if run_range is not None:
        try:
            from_run, to_run = [int(r) for r in run_range.split("-")]
        except ValueError:
            log.critical("Please specify a valid range (e.g. 3100-3200)!")
            raise SystemExit
        else:
            runs = [r for r in runs if from_run <= r.run <= to_run]

    if regex is not None:
        try:
            pattern = re.compile(regex)
        except re.error:
            log.error("Invalid regex!")
            return
        else:
            runs = [
                r
                for r in runs
                if re.match(pattern, r.runsetupname) or re.match(pattern, r.runsetupid)
            ]

    if target is not None:
        runs = [r for r in runs if r.jobtarget == target.capitalize()]

    if n is not None:
        runs = runs[-n:]

    if not runs:
        log.warning("No runs found.")
        return

    if compact:
        attributes = ["run", "runsetupname"]
        header = sep.join(attributes)

        def lineformatter(entry):
            return sep.join([str(getattr(entry, attr)) for attr in attributes])

    else:
        header = sep.join(runs[0]._fields)  # the dataset is homogenious

        def lineformatter(entry):
            return sep.join(map(str, entry))

    print(header)
    for entry in runs:
        print(lineformatter(entry))


def main():
    from docopt import docopt

    args = docopt(__doc__, version=km3db.version)

    try:
        n = int(args["-n"])
    except TypeError:
        n = None

    runtable(
        args["DET_ID"],
        n=n,
        run_range=args["-r"],
        target=args["-t"],
        regex=args["-s"],
        compact=args["-c"],
    )



import unittest
import logging
import km3db


class TestLogging(unittest.TestCase):
    def test_get_logger(self):
        log = km3db.logger.get_logger("km3db")

    def test_logs(self):
        log = km3db.logger.get_logger("km3db")
        log.debug("a debug message")
        log.info("an info message")
        log.warning("a warning")
        log.error("an error")
        log.critical("a critical error")

    def test_set_level(self):
        log = km3db.logger.get_logger("km3db")



"""
Retrieve and exploit sea current data
=====================================

The following example shows how to retrieve sea current data and
reduce them to a run per run scale.
"""

import km3db
import pandas as pd
import numpy as np

#####################################################
# Get a list of runs
# ~~~~~~~~~~~~~~~~~~
# To start, we will get a list of runs for which we later want to get
# the sea data.
#
# In this example, ORCA6 is used (det OID = "D_ORCA006") with runs 8500 to 8600

sds = km3db.tools.StreamDS(container="pd")

run_min = 8500
run_max = 8600
detid = "D_ORCA006"

runs = sds.runs(detid=detid)
runs = runs.set_index("RUN")
runs = runs[(runs.index >= run_min) & (runs.index <= run_max)]

# Add human-readable datetime object for the start and the stop of
# each run

runs["start"] = pd.to_datetime(runs["UNIXJOBSTART"], unit="ms")
runs["end"] = pd.to_datetime(runs["UNIXJOBEND"], unit="ms")

print(runs)

#####################################################
# Preparing the request
# ~~~~~~~~~~~~~~~~~~~~~
#
# The sea data are retrieved from this website:
# https://erddap.osupytheas.fr/erddap/tabledap/Emso_Ligure_Ouest_MII_Aquadopp_CSV.html
# Here we are downloading the data in csv format, then import them
# with pandas. The query is constructed in the following way:
#
# * <url>: base of the query
# * <sensor>: name of the sensor on which we query the
#   data. 'Emso_Ligure_Ouest_MII_Aquadopp_CSV' is the one close to
#   km3net ORCA.
# * <filetype>: specify the return format you want.
# * <variables>: list of variables you want to retrieve, coma
#   separated without spaces. Return everything if empty.
# * <selection>: selection criteria. In this example, we use this to
#   specify a time range.
#
# At the end, the html query should looks like :
# <url>/<sensor>.<filetype>?<variables>&<selection>

import urllib, urllib.request

# Determine the time boundary of the runs list
mindate = runs["start"].min()
maxdate = runs["end"].max()

# Prepare the query content
queryInf = {
    "url": "https://erddap.osupytheas.fr/erddap/tabledap/",
    "sensor": "Emso_Ligure_Ouest_MII_Aquadopp_CSV",
    "filetype": "csv",
    "variables": "",
    "selection": "time>={}&time<={}".format(
        mindate.strftime("%Y-%m-%dT%H:%M:%SZ"), maxdate.strftime("%Y-%m-%dT%H:%M:%SZ")
    ),
}

# Data are temporally downloaded in a csv file, in the work directory
tmp_csv_file = "tmp.csv"

# Format the query
query = "{url}{sensor}.{filetype}?".format(**queryInf)
query += urllib.request.quote("{variables}&{selection}".format(**queryInf), safe="")

# Do the query and store the returned file
urllib.request.urlretrieve(query, tmp_csv_file)

# Open the csv file with pandas.
df = pd.read_csv(open(tmp_csv_file), header=[0, 1])
# Small trick to avoid having ugly column names
df.columns = np.stack(df.columns)[:, 0]

# Add a datetime columns, based on the time string
df["datetime"] = pd.to_datetime(df["time"]).dt.tz_localize(None)

print(df)

#####################################################
# Simple plot of sea current
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To check what we have downloaded, let's do simple plots.
#
# In the first figure, we are plotting the sea current direction and
# speed, both in Cartesian coordinates (left) and polar coordinates
# (right).
#
# The left plot shows the X (pointing east) vs Y (pointing north)
# speed components of the sea current. The color scale indicate the
# angle, and shows that 0° is north, 90° is east.
#
# The right plot shows the direction vs speed in polar
# coordinates. Following left plot hints, we also tune the plot to
# display the data in the natural cardinal points orientation.

import matplotlib.pyplot as plt

fig = plt.figure(figsize=[7, 4])

ax1 = plt.subplot(121)
ax2 = plt.subplot(122, projection="polar")

ax1.set_xlabel("X east [m.s$^-1$]")
ax1.set_ylabel("Y north [m.s$^-1$]")

# Small trick to get x and y having the same scale and same boundaries
ax1.set_aspect("equal")
max_abs = 1.2 * np.max(np.abs(df[["X_East", "Y_North"]].values))
ax1.set_xlim(-max_abs, max_abs)
ax1.set_ylim(-max_abs, max_abs)

ax1.scatter(
    df["X_East"], df["Y_North"], c=df["Direction"], vmin=0, vmax=360, marker="."
)

# Replace the ticks (in rad) by cardinal points
card_angles = np.linspace(2 * np.pi, 0.0, 9)
card_labels = ["N", "NE", "E", "SE", "S", "SW", "W", "NW", "N"]
ax2.set_xticks(card_angles)
ax2.set_xticklabels(card_labels)
ax2.set_theta_zero_location("N")
ax2.set_xlabel("Sea current direction")

ax2.scatter(df["Direction"] / 180.0 * np.pi, df["Speed"], marker=".")

plt.tight_layout()

#####################################################
# Another useful thing to look at is the time evolution of the sea
# current, separately the speed and the direction :

fig, axes = plt.subplots(2, 1, sharex=True, figsize=[7, 4])

axes[0].scatter(df["datetime"], df["Direction"] / 180.0 * np.pi, marker=".")
axes[1].scatter(df["datetime"], df["Speed"], marker=".")

axes[0].set_yticks(card_angles)
axes[0].set_yticklabels(card_labels)

axes[0].set_ylabel("Direction")
axes[1].set_ylabel("Speed [m.s$^-1$]")

for ax in axes:
    ax.grid()

plt.tight_layout()


#####################################################
# Reduce to per run information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In this part, we will produce a new dataframe, that will contains
# the average value of the sea current, as well as the standard
# deviation, per data run.
#
# First, we are attaching each of the measurement to a run.


# Creating a new columns to store the run_id
df["run_id"] = np.zeros(df.shape[0], dtype=int)

for run, row in runs.iterrows():
    # Select the sea current measurements corresponding to the given run
    df_tmp = df[(df["datetime"] > row["start"]) & (df["datetime"] < row["end"])]

    # Loop over the measurements to attribute them the run_id
    for ind in df_tmp.index:
        df.at[ind, "run_id"] = run

print(df)

#####################################################
# The last columns is now showing the run_id for each measurement. Now
# let's use some pandas magic, to group the measurement per run.


df_runs = df.groupby("run_id").mean(numeric_only=True)

# the averaging process doesn't work for the angles, we therefor have
# to recompute the direction from the X and Y component of speed:
df_runs["Direction"] = np.arctan2(df_runs["X_East"], df_runs["Y_North"]) % (2 * np.pi)
df_runs["Direction"] *= 180.0 / np.pi

# Remove the orphan points
df_runs = df_runs[df_runs.index != 0]

print(df_runs)


#####################################################
# Plotting per run information
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# And finally, we want to check how look the per run information
# compare to the single measurement points.


# Determine the common index list between runs and df.
# Take the runs available in df_runs
ind = df_runs.index
# Then keep only the one also in runs
ind = ind.intersection(runs.index)


runDuration = runs.loc[ind]["end"] - runs.loc[ind]["start"]
runCenter = runs.loc[ind]["start"] + runDuration / 2.0

direction = df_runs.loc[ind]["Direction"] / 180.0 * np.pi
speed = df_runs.loc[ind]["Speed"]

fig, axes = plt.subplots(2, 1, sharex=True, figsize=[7, 4])

axes[0].scatter(
    df["datetime"], df["Direction"] / 180.0 * np.pi, marker=".", color="gray"
)
axes[1].scatter(df["datetime"], df["Speed"], marker=".", color="gray")

axes[0].errorbar(runCenter, direction, xerr=runDuration / 2.0, fmt=".")
axes[1].errorbar(runCenter, speed, xerr=runDuration / 2.0, fmt=".")

axes[0].set_yticks(card_angles)
axes[0].set_yticklabels(card_labels)

axes[0].set_ylabel("Direction")
axes[1].set_ylabel("Speed [m.s$^-1$]")

for ax in axes:
    ax.grid()

plt.tight_layout()
plt.show()



#!/usr/bin/env python3


def pandas():
    """Imports and returns ``pandas``."""
    try:
        import pandas
    except ImportError:
        raise ImportError(
            "install the 'pandas' package with:\n\n"
            "    pip install pandas\n\n"
            "or\n\n"
            "    conda install pandas"
        )
    else:
        return pandas


def h5py():
    """Imports and returns ``h5py``."""
    try:
        import h5py
    except ImportError:
        raise ImportError(
            "install the 'h5py' package with:\n\n"
            "    pip install h5py\n\n"
            "or\n\n"
            "    conda install h5py"
        )
    else:
        return h5py



#!/usr/bin/env python3
from hashlib import sha256
import os
import re
import sys
import logging
import logging.handlers

loggers = {}  # this holds all the registered loggers


ATTRIBUTES = dict(
    list(
        zip(
            ["bold", "dark", "", "underline", "blink", "", "reverse", "concealed"],
            list(range(1, 9)),
        )
    )
)
del ATTRIBUTES[""]

ATTRIBUTES_RE = r"\033\[(?:%s)m" % "|".join(["%d" % v for v in ATTRIBUTES.values()])

HIGHLIGHTS = dict(
    list(
        zip(
            [
                "on_grey",
                "on_red",
                "on_green",
                "on_yellow",
                "on_blue",
                "on_magenta",
                "on_cyan",
                "on_white",
            ],
            list(range(40, 48)),
        )
    )
)

HIGHLIGHTS_RE = r"\033\[(?:%s)m" % "|".join(["%d" % v for v in HIGHLIGHTS.values()])

COLORS = dict(
    list(
        zip(
            [
                "grey",
                "red",
                "green",
                "yellow",
                "blue",
                "magenta",
                "cyan",
                "white",
            ],
            list(range(30, 38)),
        )
    )
)

COLORS_RE = r"\033\[(?:%s)m" % "|".join(["%d" % v for v in COLORS.values()])

RESET = r"\033[0m"
RESET_RE = r"\033\[0m"


def supports_color():
    """Checks if the terminal supports color."""
    if isnotebook():
        return True
    supported_platform = sys.platform != "win32" or "ANSICON" in os.environ
    is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    if not supported_platform or not is_a_tty:
        return False

    return True


def colored(text, color=None, on_color=None, attrs=None, ansi_code=None):
    """Colorize text, while stripping nested ANSI color sequences.

    Author:  Konstantin Lepa <konstantin.lepa@gmail.com> / termcolor

    Available text colors:
        red, green, yellow, blue, magenta, cyan, white.
    Available text highlights:
        on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.
    Available attributes:
        bold, dark, underline, blink, reverse, concealed.
    Example:
        colored('Hello, World!', 'red', 'on_grey', ['blue', 'blink'])
        colored('Hello, World!', 'green')
    """
    if os.getenv("ANSI_COLORS_DISABLED") is None:
        if ansi_code is not None:
            return "\033[38;5;{}m{}\033[0m".format(ansi_code, text)
        fmt_str = "\033[%dm%s"
        if color is not None:
            text = re.sub(COLORS_RE + "(.*?)" + RESET_RE, r"\1", text)
            text = fmt_str % (COLORS[color], text)
        if on_color is not None:
            text = re.sub(HIGHLIGHTS_RE + "(.*?)" + RESET_RE, r"\1", text)
            text = fmt_str % (HIGHLIGHTS[on_color], text)
        if attrs is not None:
            text = re.sub(ATTRIBUTES_RE + "(.*?)" + RESET_RE, r"\1", text)
            for attr in attrs:
                text = fmt_str % (ATTRIBUTES[attr], text)
        return text + RESET
    else:
        return text


def hash_coloured_escapes(text):
    """Return the ANSI hash colour prefix and suffix for a given text"""
    ansi_code = int(sha256(text.encode("utf-8")).hexdigest(), 16) % 230
    prefix, suffix = colored("SPLIT", ansi_code=ansi_code).split("SPLIT")
    return prefix, suffix


def isnotebook():
    """Check if running within a Jupyter notebook"""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False


def get_logger(name, filename=None, stream_loglevel="INFO", file_loglevel="DEBUG"):
    """Helper function to get a logger"""
    if name in loggers:
        return loggers[name]
    logger = logging.getLogger(name)
    logger.propagate = False

    with_color = supports_color()

    pre1, suf1 = hash_coloured_escapes(name) if with_color else ("", "")
    pre2, suf2 = hash_coloured_escapes(name + "salt") if with_color else ("", "")
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s {}+{}+{} "
        "%(name)s: %(message)s".format(pre1, pre2, suf1),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    if filename is not None:
        ch_file = logging.handlers.RotatingFileHandler(
            filename, maxBytes=5 * 1024 * 1024, backupCount=10
        )
        ch_file.setLevel(file_loglevel)
        ch_file.setFormatter(formatter)
        logger.addHandler(ch_file)
    ch = logging.StreamHandler()
    ch.setLevel(stream_loglevel)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    loggers[name] = logger

    logger.once_dict = {}

    return logger


def set_level(log_or_name, level):
    """Set the log level for given logger"""
    if isinstance(log_or_name, str):
        log = get_logger(log_or_name)
    else:
        log = log_or_name
    log.setLevel(level)
    for handler in log.handlers:
        handler.setLevel(level)


if supports_color():
    logging.addLevelName(
        logging.INFO, "\033[1;32m%s\033[1;0m" % logging.getLevelName(logging.INFO)
    )
    logging.addLevelName(
        logging.DEBUG, "\033[1;34m%s\033[1;0m" % logging.getLevelName(logging.DEBUG)
    )
    logging.addLevelName(
        logging.WARNING, "\033[1;33m%s\033[1;0m" % logging.getLevelName(logging.WARNING)
    )
    logging.addLevelName(
        logging.ERROR, "\033[1;31m%s\033[1;0m" % logging.getLevelName(logging.ERROR)
    )
    logging.addLevelName(
        logging.CRITICAL,
        "\033[1;101m%s\033[1;0m" % logging.getLevelName(logging.CRITICAL),
    )


log = get_logger("km3db")
set_level(log, "WARNING")



#!/usr/bin/env python
# Filename: setup.py
"""
The km3db setup script.

"""
from setuptools import setup
import os
import sys


def read_requirements(kind):
    """Return a list of stripped lines from a file"""
    with open(os.path.join("requirements", kind + ".txt")) as fobj:
        requirements = [l.strip() for l in fobj.readlines()]
    v = sys.version_info
    if (v.major, v.minor) < (3, 6):
        try:
            requirements.pop(requirements.index("black"))
        except ValueError:
            pass
    return requirements


try:
    with open("README.rst") as fh:
        long_description = fh.read()
except UnicodeDecodeError:
    long_description = "KM3NeT database library"

setup(
    name="km3db",
    url="https://git.km3net.de/km3py/km3db",
    description="KM3NeT database library",
    long_description=long_description,
    author="Tamas Gal",
    author_email="tgal@km3net.de",
    packages=["km3db"],
    include_package_data=True,
    platforms="any",
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    python_requires=">=2.7",
    install_requires=read_requirements("install"),
    extras_require={kind: read_requirements(kind) for kind in ["dev", "extras"]},
    entry_points={
        "console_scripts": [
            "km3db=km3db.cli.km3db:main",
            "km3dbcookie=km3db.cli.km3dbcookie:main",
            "streamds=km3db.cli.streamds:main",
            "detx=km3db.cli.detx:main",
            "runtable=km3db.cli.runtable:main",
            "runinfo=km3db.cli.runinfo:main",
            "wtd=km3db.cli.wtd:main",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
    ],
)



#!/usr/bin/env python3
# Filename: core.py
"""
Database utilities.

"""
from __future__ import absolute_import, print_function, division

import ssl
import getpass
import os
import re
from urllib.error import URLError, HTTPError
from http.client import IncompleteRead, RemoteDisconnected
import socket
import pytz
import time
from urllib.parse import unquote
import urllib.request

from km3db.logger import log


BASE_URL = "https://km3netdbweb.in2p3.fr"
COOKIE_FILENAME = os.path.expanduser("~/.km3netdb_cookie")
SESSION_COOKIES = dict(
    gitlab="_gitlab-km3net_131.188_ce0e106433dd4923b522716b23c992c2",
)
UTC_TZ = pytz.timezone("UTC")

_cookie_sid_pattern = re.compile(r"_[a-z0-9-]+_(\d{1,3}.){1,3}\d{1,3}_[a-z0-9]+")

# Ignore invalid certificate error
ssl._create_default_https_context = ssl._create_unverified_context

original_getaddrinfo = socket.getaddrinfo


def ipv4_forced_getaddrinfo(*args, **kwargs):
    responses = original_getaddrinfo(*args, **kwargs)
    return [res for res in responses if res[0] == socket.AF_INET]


# Monkey patch to force IPv4
socket.getaddrinfo = ipv4_forced_getaddrinfo


class AuthenticationError(Exception):
    pass


class DBManager:
    def __init__(self, url=None, network_class=None):
        self._db_url = BASE_URL if url is None else url
        self._login_url = self._db_url + "/home.htm"
        self._network_class = network_class
        self._session_cookie = None
        self._opener = None
        self._username = None

    def get(self, url, default=None, retries=10):
        "Get HTML content"
        target_url = self._db_url + "/" + unquote(url)
        log.debug("Accessing %s", target_url)
        try:
            f = self.opener.open(target_url)
        except HTTPError as e:
            if e.code == 403:
                if retries:
                    log.error(
                        "Access forbidden, your session has expired. "
                        "Deleting the cookie (%s) and retrying once.",
                        COOKIE_FILENAME,
                    )
                    retries -= 1
                else:
                    log.critical("Access forbidden. Giving up...")
                    return default
                time.sleep(1)
                self.reset()
                if os.path.exists(COOKIE_FILENAME):
                    os.remove(COOKIE_FILENAME)
                return self.get(url, default=default, retries=retries)
            log.error("HTTP error: %s\n" "Target URL: %s", e, target_url)
            return default
        except URLError as e:
            if retries:
                retries -= 1
                log.error("URLError '%s', retrying in 30 seconds.", e)
                time.sleep(30)
                return self.get(url, default=default, retries=retries)
            else:
                log.error("Giving up... URLError: %s\n" "Target URL: %s", e, target_url)
                return default
        except RemoteDisconnected as e:
            if retries:
                retries -= 1
                log.error("RemoteDisconnected '%s', retrying in 30 seconds.", e)
                time.sleep(30)
                return self.get(url, default=default, retries=retries)
            else:
                log.error(
                    "Giving up... RemoteDisconnected: %s\n" "Target URL: %s",
                    e,
                    target_url,
                )
                return default
        try:
            content = f.read()
        except IncompleteRead as icread:
            log.error("Incomplete data received from the DB.")
            content = icread.partial
        log.debug("Got {0} bytes of data.".format(len(content)))

        return content.decode("utf-8")

    def reset(self):
        "Reset everything"
        self._opener = None
        self._session_cookie = None

    @property
    def session_cookie(self):
        if self._session_cookie is None:
            for host, session_cookie in SESSION_COOKIES.items():
                if on_whitelisted_host(host):
                    self._session_cookie = session_cookie
                    break
            else:
                self._session_cookie = self._request_session_cookie()
        return self._session_cookie

    def _request_session_cookie(self):
        """Request cookie for permanent session."""
        # The cookie can be specified via the environment
        cookie = os.getenv("KM3NET_DB_COOKIE")
        if cookie is not None:
            log.info("Using cookie from env ($KM3NET_DB_COOKIE)")
            return cookie

        # The cookie file can also be specified via the environment
        cookiefile = os.getenv("KM3NET_DB_COOKIE_FILE")
        if cookiefile is not None:
            log.info("Using cookie file from env ($KM3NET_DB_COOKIE_FILE)")
            with open(cookiefile) as fobj:
                content = fobj.read()
            return content.split()[-1].strip()

        # Next, try the configuration file according to
        # the specification described here:
        # https://wiki.km3net.de/index.php/Database#Scripting_access
        if os.path.exists(COOKIE_FILENAME):
            log.info("Using cookie from standard location %s", COOKIE_FILENAME)
            # TODO: code duplication, see above
            with open(COOKIE_FILENAME) as fobj:
                content = fobj.read()
            return content.split()[-1].strip()

        # If everything fails, ask the user for credentials to generate a cookie
        return self.request_session_cookie()

    def request_session_cookie(self):
        """Request a cookie using credentials"""
        username = os.getenv("KM3NET_DB_USERNAME")
        password = os.getenv("KM3NET_DB_PASSWORD")

        if username is None or password is None:
            # Last resort: we ask interactively
            username = input("Please enter your KM3NeT DB username: ")
            password = getpass.getpass("Password: ")
        else:
            log.info(
                "Using credentials from env ($KM3NET_DB_USERNAME and "
                "$KM3NET_DB_PASSWORD)"
            )

        suffix = ""
        if self._network_class is not None:
            if self._network_class == "B":
                suffix = "&freenetbits=16"
            elif self._network_class == "C":
                suffix = "&freenetbits=8"
            else:
                log.error("Unsupported network class '{}'".format(self._network_class))
        target_url = self._login_url + "?usr={0}&pwd={1}&persist=y{2}".format(
            username, password, suffix
        )
        cookie = urllib.request.urlopen(target_url).read()

        # Unicode madness
        try:
            cookie = str(cookie, "utf-8")  # Python 3
        except TypeError:
            cookie = str(cookie)  # Python 2

        cookie = cookie.split("sid=")[-1]

        if not _cookie_sid_pattern.match(cookie):
            message = "Wrong username or password."
            log.critical(message)
            raise AuthenticationError(message)

        log.info("Writing session cookie to %s", COOKIE_FILENAME)
        with open(COOKIE_FILENAME, "w") as fobj:
            fobj.write(".in2p3.fr\tTRUE\t/\tTRUE\t0\tsid\t{}".format(cookie))

        return cookie

    @property
    def opener(self):
        "A reusable connection manager"
        if self._opener is None:
            log.debug("Creating connection handler")
            opener = urllib.request.build_opener()
            cookie = self.session_cookie
            if cookie is None:
                log.critical("Could not connect to database.")
                return
            opener.addheaders.append(("Cookie", "sid=" + cookie))
            log.debug("Using session cookie: sid=%s", cookie)
            self._opener = opener
        else:
            log.debug("Reusing connection manager")
        return self._opener

    @property
    def username(self):
        if self._username is None:
            self._username = self.session_cookie.split("_")[1]
        return self._username


def on_whitelisted_host(name):
    """Check if we are on a whitelisted host"""
    if name == "gitlab":
        return "GITLAB_CI" in os.environ



# Filename: runinfo.py
"""
Prints the run table for a given detector ID.

Usage:
    runinfo DET_ID RUN
    runinfo (-h | --help)
    runinfo --version

Options:
    -h --help           Show this screen.
    DET_ID              Detector ID (eg. D_ARCA001).
    RUN                 Run number.

"""
from datetime import datetime, timezone
import km3db

log = km3db.logger.log


def runinfo(run_id, det_id):
    runs = km3db.StreamDS(container="nt").get("runs", detid=det_id)

    if runs is None:
        log.error("No runs found for detector ID {}".format(det_id))
        return

    for idx, run in enumerate(runs):
        if run.run == run_id:
            break
    else:
        log.error("No run with ID {} found for detector ID".format(run_id, det_id))
        return

    try:
        next_run = runs[idx + 1]
    except IndexError:
        next_run = None
        end_time = duration = float("NaN")
    else:
        duration = (next_run.unixstarttime - run.unixstarttime) / 1000 / 60
        end_time = iso8601utc(next_run.unixstarttime / 1000)

    print("Run {0} - detector ID: {1}".format(run_id, det_id))
    print("-" * 42)
    print(
        "  Start time:         {0}\n"
        "  End time:           {1}\n"
        "  Duration [min]:     {2:.2f}\n"
        "  Start time defined: {3}\n"
        "  Runsetup ID:        {4}\n"
        "  Runsetup name:      {5}\n"
        "  T0 Calibration ID:  {6}".format(
            iso8601utc(run.unixstarttime / 1000),
            end_time,
            duration,
            bool(run.starttime_defined),
            run.runsetupid,
            run.runsetupname,
            run.t0_calibsetid,
        )
    )


def iso8601utc(timestamp):
    return datetime.fromtimestamp(timestamp, tz=timezone.utc).isoformat()


def main():
    from docopt import docopt

    args = docopt(__doc__, version=km3db.version)

    runinfo(int(args["RUN"]), int(args["DET_ID"]))
